#include <sys/poll.h>
#include <time.h>
#include <syslog.h>
#include <sys/stat.h>
#include <signal.h>
#include <stdint.h>
#include <stdlib.h>
#include "headers.h"
#include "network.h"
//#include "utils.h"
#include "uart.h"
#include "protocol.h"


char *INTERFACE_NAME = "enp4s0";

char *PID_FILE_NAME = "/var/run/mc0203_daemon.pid";

typedef struct {
  char *uart_name;
  int boudrate;
  char *ifname;
} ctx_t;

ctx_t ctx;
// returns 0 on success -1 on error
int become_daemon();

int create_pid_file(char *filename);

void read_cmd_args(int argc, char *argv[]);

//
int check_daemon();

int exec(struct uart_t *self);

void sigintHandler(int sig_num) {
    char log[256];
    snprintf(log, 256, "            DAEMON ENDED - errno = %d - SIGNAL = %d", errno, sig_num);
    syslog(LOG_DEBUG, log, strlen(log));
    remove(PID_FILE_NAME);
    exit(0);
}

int main(int argc, char *argv[]) {
    char log_message[256];
    ctx.uart_name = "/dev/ttyUSB0";
    ctx.ifname = "eth0";
    ctx.boudrate = 115200;

    openlog("mc03_uart", LOG_PID, LOG_DAEMON);
    read_cmd_args(argc, argv);
    syslog(LOG_INFO, "start mc03_uart");

     if (check_daemon()) {
         snprintf(log_message, 256, "DAEMON IS ALREADY RUNNING");
         syslog(LOG_DEBUG, log_message, strlen(log_message));
         exit(0);
     }
     if (become_daemon() != 0) {
         return EXIT_FAILURE;
     }
     create_pid_file(PID_FILE_NAME);
     signal(SIGTERM, sigintHandler);

    syslog(LOG_INFO, "APP STARTED");
    sprintf(log_message, "Network interface - %s", INTERFACE_NAME);
    syslog(LOG_INFO, log_message, strlen(log_message));
    sprintf(log_message, "UART - %s", UART_NAME);
    syslog(LOG_INFO, log_message, strlen(log_message));
    sprintf(log_message, "Baudrate - %d", BAUDRATE);
    syslog(LOG_INFO, log_message, strlen(log_message));

    uint32_t *ip = NULL;
    /* printf("try get memmory"); */
    ip = (uint32_t*) malloc(sizeof(uint32_t));
    get_iface_ip(ctx.ifname, ip);
    uint8_t *i;
    i = (uint8_t*) ip;
    printf("get ip addr %d.%d.%d.%d\n", i[0],i[1],i[2],i[3]);

    get_iface_mask(ctx.ifname, ip);
    i = (uint8_t*) ip;
    printf("get ip addr %d.%d.%d.%d\n", i[0],i[1],i[2],i[3]);

    get_default_gw(ip);
    i = (uint8_t*) ip;
    printf("get ip addr %d.%d.%d.%d\n", i[0],i[1],i[2],i[3]);

    struct uart_t main_uart;
    memset(&main_uart, 0, sizeof(struct uart_t));
    main_uart.invoke = exec;
    printf("open uart\n");
    //init_uart(&main_uart, ctx.uart_name, ctx.boudrate);
     if (init_uart(&main_uart, ctx.uart_name, ctx.boudrate) == -1) {
         sprintf(log_message, "%s: %s", "UART ERROR, app crashed", strerror(errno));
         syslog(LOG_ERR, log_message, strlen(log_message));
         sigintHandler(0);
         exit(1);
     }
    printf("open success uart\n");
    /* int d = init_network_interface(INTERFACE_NAME); */
    /* if (d != 0) { */
    /*     sprintf(log_message, "%s: %s", "NETWORK ERROR, app crashed", strerror(errno)); */
    /*     syslog(LOG_ERR, log_message, strlen(log_message)); */
    /*     sigintHandler(0); */
    /*     exit(1); */
    /* } */
    syslog(LOG_DEBUG, "test");
    struct pollfd uart_poll;
    uart_poll.fd = main_uart.fd;
    uart_poll.events = POLLIN;
    const int TIMEOUT = 100000;  // Timeout in milliseconds


    syslog(LOG_DEBUG, "start poll message");
    while (1) {
      printf("read cicle\n");
        int ret = poll(&uart_poll, 1, TIMEOUT);
        if (ret == -1) {
            perror("Failed to poll UART device");
            break;
        } else if (ret == 0) {
            // Timeout occurred
            continue;
        }
        if (uart_poll.revents & POLLIN) {
            read_uart_circle(&main_uart);
        }
    }


}

void read_cmd_args(int argc, char *argv[]) {
    char log[256];

    for (int i = 1; i < argc; i++) {
        if (strcmp(argv[i], "-b") == 0 && i + 1 < argc) {
            ctx.boudrate = atoi(argv[++i]);
        } else {
            if (strcmp(argv[i], "-u") == 0 && i + 1 < argc) {
                ctx.uart_name = argv[++i];
            } else {
                if (strcmp(argv[i], "-n") == 0 && i + 1 < argc) {
                    ctx.ifname = argv[++i];
                } else {
                    if (strcmp(argv[i], "--help") == 0) {
                        snprintf(log, 256, "usage %s:\n%s\n%s\n%s\n",
                                 argv[0],
                                 "-n - specify network interface name",
                                 "-u - specify uart interface name",
                                 "-b - specify uart baudrate");
                        printf("%s\n",log);
                        exit(0);
                    } else {
                        snprintf(log, 256, "%s: invalid option -- '%s'\nTry '%s --help' for more information.\n",
                                 argv[0],
                                 argv[i],
                                 argv[0]);
                        syslog(LOG_ERR, log, strlen(log));
                        printf("%s\n",log);
                        exit(0);
                    }
                }
            };

        }
    }
}

int exec(struct uart_t *self) {
    uint8_t index = 0xad;
    struct config_t config;
    char data_buffer[256], log_buffer[256];
    memset(data_buffer, 0, 256);

    int data_size = 0;
    int command;
    command = extract_data_from_command(self->message, 256, data_buffer, &data_size);
    switch (command) {
        case COMMAND_SET_CONFIG : {
            printf("\n\nCOMMAND_SET_CONFIG\n\n");

            convert_from_string(data_buffer, &(config.dhcp), &(config.ip), &(config.netmask), &(config.gw),
                                &(config.dns));
            snprintf(log_buffer, 300, "SET CONFIG - %s", data_buffer);
            syslog(LOG_DEBUG, log_buffer, strlen(log_buffer));
            //char flag = set_c(&(config.dhcp), &(config.ip), &(config.netmask), &(config.gw),
            //                  &(config.dns), INTERFACE_NAME);
            int flag = set_iface_mask(ctx.ifname, config.netmask);
            set_default_gw(config.gw);
            set_dns(config.dns);
            if (flag != 0) {
                sprintf((char *)data_buffer, "%s: %s", "SET_CONFIG ERROR - ", strerror(errno));
                syslog(LOG_ERR, data_buffer, strlen(data_buffer));

            }
            //char *msg;
            //msg = convert_from_string(msg, config);
            get_response(data_buffer, &data_size, (char *)&flag, 0, COMMAND_SET_CONFIG_RESPONSE, index);
            for (int i = 0; i < data_size; ++i) {
                printf("%03d ", data_buffer[i]);
            }
            write(self->fd, data_buffer, data_size);
            break;
        }
        case COMMAND_GET_CONFIG: {
            memset(self->message, 0, 200);
            printf("\n\nCOMMAND_GET_CONFIG\n\n");
            /* get_c(&config.dhcp, &(config.ip), &(config.netmask), &(config.gw), &(config.dns)); */
            get_iface_ip(ctx.ifname, &(config.ip));
            get_iface_mask(ctx.ifname, &(config.netmask));
            get_default_gw(&(config.gw));
            config.dhcp = 0;
            get_dns(&(config.dns));
            convert_to_string(self->message, &(config));

            get_response(data_buffer,
                &data_size,
                self->message,
                200,
                COMMAND_GET_CONFIG_RESPONSE,
                index);
            syslog(LOG_DEBUG, "DATASIZE - %d RESPONSE FOR GET - ", data_size);
            for (int i = 0; i <= data_size; i++) {
                printf("%03d ", data_buffer[i]);
            }
            write(self->fd, data_buffer, data_size - 1);
            break;
        }
    }
    return 0;
}

int become_daemon() {
  syslog(LOG_INFO, "try start daemon");
    switch (fork()) {
        case -1:
            return -1;
        case 0:
            break;
        default:
            _exit(EXIT_SUCCESS);
    }

    if (setsid() == -1)
        return -1;

    switch (fork()) {
        case -1:
            return -1;
        case 0:
            break;
        default:
            _exit(EXIT_SUCCESS);
    }

    chdir("/");
    close(STDIN_FILENO);

    int fd = open("/dev/null", O_RDWR);
    if (fd != STDIN_FILENO)
        return -1;
    if (dup2(STDIN_FILENO, STDOUT_FILENO) != STDOUT_FILENO)
        return -2;
    if (dup2(STDIN_FILENO, STDERR_FILENO) != STDERR_FILENO)
        return -3;
    return 0;
}

int check_daemon() {
    int qwerty = open(PID_FILE_NAME, O_RDONLY);
    if (qwerty != -1) {
        close(qwerty);
        return 1;
    } else {
        close(qwerty);
        return 0;
    }
}

int create_pid_file(char *filename) {
    int qwerty = open(filename, O_CREAT | O_WRONLY);
    char buff[10];
    snprintf(buff, 10, "%d", getpid());
    write(qwerty, buff, strlen(buff));
    close(qwerty);
    return 0;
}


